/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.hayabusa.db;

import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.fukurou.util.LogWriter;
import org.opengion.fukurou.util.HybsDateUtil;

/**
 * データのコード情報を取り扱うクラスです。
 *
 * パラメータで指定した、開始、終了、ステップ、フォーマット、titleフォーマットの情報から、
 * HTMLのメニューやリストを作成するための オプション
 * タグを作成したり、与えられたキーをもとに、チェック済みのオプションタグを作成したりします。
 * パラメータの初期値は、開始(0)、終了(14)、ステップ(1)、フォーマット(MM/dd)、titleフォーマット(null) です。
 * 
 * ここでは、日付(年/月/日）の自動生成を行います。
 * 基準は、実行日に対して、＋－の数字を指定します。
 * 例：20130206 に対して、-2,3,1 を指定すると、20130204,20130205,20130206,20130207,20130208,20130209 となります。
 * 
 * 開始、終了に、特殊なコマンドを指定する事も可能です。
 * ・SD ：当月の最初の日付にセットします。(当月１日)
 * ・ED ：当月の最後の日付にセットします。(当月月末)
 * ・SW ：日付処理の週初め(月曜日)にセットします。日付は当日より前に移動します。
 * ・EW ：日付処理の週末(日曜日)にセットします。日付は当日より後ろに移動します。
 * ・D1 ～ DXXX ：日を指定の分だけ進めます。D1なら翌日、D200 なら200日後
 * ・M1 ～ MXXX ：月を指定の分だけ進めます。M1なら翌月、M6 なら半年後
 * ・BSD ：先月の最初の日付にセットします。(先月１日)
 * ・BED ：先月の最後の日付にセットします。(先月月末)
 * ・ASD ：翌月の最初の日付にセットします。(翌月１日)
 * ・AED ：翌月の最後の日付にセットします。(翌月月末)
 * ・M1  ：１か月先。数字部分は、任意に指定できるため、M5 とすれば、５か月先。M-2 は、２か月前
 *
 * ステップは、無指定の場合は、１日単位です。例えば、SW,M6,7 とすれば、毎週月曜日を６か月先まで作成します。
 * また、M1 は、１か月単位になります。ただし、開始日を SD などにしないと、大の月と小の月で、おかしくなるので
 * ご注意ください。
 *
 * キーは、８文字の yyyyMMdd 形式で与えられます。ラベルは、フォーマットの指定に準拠します。
 * フォーマットを指定する場合は、ステップは、必須となります。
 * フォーマットの初期値は、MM/dd です。
 * titleフォーマット は、tips表示する場合のフォーマットになります。初期値は、ありません。
 *
 * @og.group 選択データ制御
 * @og.rev 5.6.1.1 (2013/02/08) 新規追加
 *
 * @version  4.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public class Selection_YMD implements Selection {
	private final String   CACHE ;
	private final String   ST_ED_STEP ;

	private final DateFormat lblFmt  ;
	private final DateFormat tipsFmt ;

	private final long	maxCacheTime ;		// キャッシュの破棄タイミングを計るための最大有効時間

	/**
	 * コンストラクター
	 *
	 * 引数は、開始、終了、ステップ、フォーマット、titleフォーマットです。
	 * パラメータの初期値は、開始(0)、終了(14)、ステップ(1)、フォーマット(MM/dd)、titleフォーマット(null) です。
	 *
	 * @param	editPrm	開始、終了、ステップ、フォーマット、titleフォーマットを表す引数(-2,3,1)
	 */
	public Selection_YMD( final String editPrm ) {
	//	if( param.length < 2 ) {
	//		String errMsg = "引数は、開始、終了、[ステップ]、[フォーマット]、 [titleフォーマット]です。最低でも２個必要です。";
	//		throw new IllegalArgumentException( errMsg );
	//	}

		String[] param = (editPrm == null) ? new String[0] : editPrm.split( "," ) ;

		String start   = (param.length > 0) ? param[0].trim() : "0" ;
		String end     = (param.length > 1) ? param[1].trim() : "14" ;

		String step    = (param.length > 2) ? param[2].trim() : "1" ;
		String lblPrm  = (param.length > 3) ? param[3].trim() : "MM/dd" ;
		String tipsPrm = (param.length > 4) ? param[4].trim() : "MM月dd日(EEE)" ;

		Calendar cal = Calendar.getInstance();
		HybsDateUtil.calendarCalc( cal , start );

		Calendar endCal = Calendar.getInstance();
		HybsDateUtil.calendarCalc( endCal , end );

		DateFormat keyFmt  = new SimpleDateFormat( "yyyyMMdd",Locale.JAPAN );					// キーとなるフォーマット。yyyyMMdd 固定
		lblFmt  = new SimpleDateFormat( lblPrm,Locale.JAPAN );									// ラベルのフォーマット
		tipsFmt = ( tipsPrm != null ) ? new SimpleDateFormat( tipsPrm,Locale.JAPAN ) : null ;	// titleフォーマット

		ST_ED_STEP = "Start=" + lblFmt.format( cal.getTime() ) +
					 " , End=" + lblFmt.format( endCal.getTime() ) + " , Step=" + step ;

		StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_MIDDLE );

		// cal.before( endCal ) では、同一時刻の場合に false になる為、ここの判定では使えません。
		// sign を掛け算しているのは、逆順対応
		int sign = ( step.indexOf( '-' ) < 0 ) ? 1 : -1 ;	// ステップの符号。
		while( endCal.compareTo( cal ) * sign >= 0 ) {
			String key  = keyFmt.format( cal.getTime() );
			String val  = lblFmt.format( cal.getTime() );

			buf.append( "<option value=\"" ).append( key ).append( "\"" );
			if( tipsFmt != null ) {
				String tips = tipsFmt.format( cal.getTime() ) ;
				buf.append(" title=\"").append( tips ).append("\"");
			}
			buf.append( ">" ).append( val ).append( "</option>" );

			HybsDateUtil.calendarCalc( cal , step );
		}

		CACHE = buf.toString();

		// キャシュの有効期間を求めるための時刻を作成します。キャッシュは、当日のみ有効です。
		Calendar now   = Calendar.getInstance();
		now.set( Calendar.HOUR   , 0 );				// 時、分、秒 をリセットします。
		now.set( Calendar.MINUTE , 0 );
		now.set( Calendar.SECOND , 0 );
		now.add( Calendar.DAY_OF_MONTH , 1 );		// 1日進めます。

		maxCacheTime = now.getTimeInMillis() ;
	}

	/**
	 * 初期値が選択済みの 選択肢(オプション)を返します。
	 * このオプションは、引数の値を初期値とするオプションタグを返します。
	 * このメソッドでは、ラベル(短)が設定されている場合でも、これを使用せずに必ずラベル(長)を使用します。
	 *
	 * @og.rev 3.6.0.6 (2004/10/22) シーケンスアクセス機能を指定する seqFlag を導入
	 * @og.rev 4.0.0 (2005/01/31) selectValue が、null/ゼロ文字列でない場合に、選択肢にない場合は、エラー
	 *
	 * @param   selectValue  選択されている値
	 * @param   seqFlag  シーケンスアクセス機能 [true:ON/false:OFF]
	 *
	 * @return  オプションタグ
	 * @see     #getOption( String, boolean, boolean )
	 */
	public String getOption( final String selectValue,final boolean seqFlag ) {
		return getOption( selectValue, seqFlag, false );
	}

	/**
	 * 初期値が選択済みの 選択肢(オプション)を返します。
	 * このオプションは、引数の値を初期値とするオプションタグを返します。
	 * このメソッドでは、引数のuseShortLabelがtrueに指定された場合に、ラベル(短)をベースとした
	 * ツールチップ表示を行います。
	 *
	 * @param   selectValue  選択されている値
	 * @param   seqFlag  シーケンスアクセス機能 [true:ON/false:OFF]
	 * @param   useShortLabel ラベル(短)をベースとしたオプション表示を行うかどうか。(未使用)
	 *
	 * @return  オプションタグ
	 * @see     #getOption( String, boolean )
	 */
	public String getOption( final String selectValue,final boolean seqFlag, final boolean useShortLabel ) {
		// マッチするアドレスを探す。
		int selected = CACHE.indexOf( "\"" + selectValue + "\"" );

		if( selected < 0 ) {
			if( selectValue != null && selectValue.length() > 0 ) {
				String errMsg = "年月日範囲に存在しない値が指定されました。"
							+ " value=[" + selectValue + "]"
							+ HybsSystem.CR + ST_ED_STEP ;
				LogWriter.log( errMsg );
			}
			return CACHE;
		}
		else {
			// "年月日" 文字列の位置が、selected なので、年月日の文字数＋２までが、前半部分になる。(年月日の文字数は８固定のはず)
			int indx = selected + selectValue.length() + 2 ;

			StringBuilder buf = new StringBuilder( HybsSystem.BUFFER_MIDDLE );
			// 3.6.0.6 (2004/10/22) シーケンスアクセス機能を指定する seqFlag を導入
			if( seqFlag ) {
				buf.append( "<option value=\"" ).append( selectValue ).append( "\"" );
			}
			else {
				buf.append( CACHE.substring( 0,indx ) );
			}
			buf.append( " selected=\"selected\"" );
			buf.append( CACHE.substring( indx ) );
			return buf.toString() ;
		}
	}

	/**
	 * 初期値が選択済みの 選択肢(オプション)を返します。
	 * このオプションは、引数の値を初期値とするオプションタグを返します。
	 * ※ このクラスでは実装されていません。
	 *
	 * @og.rev 2.1.0.1 (2002/10/17) 選択リストを、正方向にしか選べないようにする sequenceFlag を導入する
	 * @og.rev 3.8.6.0 (2006/09/29) useLabel 属性 追加
	 *
	 * @param   name         ラジオの name
	 * @param   selectValue  選択されている値
	 * @param   useLabel     ラベル表示の有無 [true:有/false:無]
	 *
	 * @return  オプションタグ
	 */
	public String getRadio( final String name,final String selectValue,final boolean useLabel ) {
		String errMsg = "このクラスでは実装されていません。";
		throw new UnsupportedOperationException( errMsg );
	}

	/**
	 * 初期値が選択済みの 選択肢(オプション)を返します。
	 * このオプションは、引数の値を初期値とするオプションタグを返します。
	 * ※ このクラスでは実装されていません。
	 *
	 * @param   selectValue  選択されている値
	 *
	 * @return  オプションタグ
	 */
	public String getRadioLabel( final String selectValue ) {
		String errMsg = "このクラスでは実装されていません。";
		throw new UnsupportedOperationException( errMsg );
	}

	/**
	 * 選択肢(value)に対するラベルを返します。
	 * 選択肢(value)が、存在しなかった場合は、選択肢そのものを返します。
	 * getValueLabel( XX ) は、getValueLabel( XX,false ) と同じです。
	 *
	 * @param   selectValue 選択肢の値
	 *
	 * @return  選択肢のラベル
	 * @see     #getValueLabel( String,boolean )
	 */
	public String getValueLabel( final String selectValue ) {
		return getValueLabel( selectValue,false );
	}

	/**
	 * 選択肢(value)に対するラベルを返します。
	 * 選択肢(value)が、存在しなかった場合は、選択肢そのものを返します。
	 * このメソッドでは、短縮ラベルを返すかどうかを指定するフラグを指定します。
	 * getValueLabel( XX,false ) は、getValueLabel( XX ) と同じです。
	 *
	 * @og.rev 4.0.0 (2005/11/30) を追加
	 *
	 * @param	selectValue	選択肢の値
	 * @param	flag	短縮ラベルを [true:使用する/false:しない]
	 *
	 * @return  選択肢のラベル
	 * @see     #getValueLabel( String )
	 */
	public String getValueLabel( final String selectValue,final boolean flag ) {
		// 選択肢があろうとなかろうと、フォーマット変換して値を返す。

		Calendar cal = HybsDateUtil.getCalendar( selectValue );

		if( flag && tipsFmt != null ) {
			return "<span title=\"" + tipsFmt.format( cal.getTime() ) + "\">" + lblFmt.format( cal.getTime() ) + "</span>";
		}
		else {
			return lblFmt.format( cal.getTime() );
		}
	}

	/**
	 * マルチ・キーセレクトを使用するかどうかを返します。
	 * true：使用する。false:使用しない です。
	 * ただし、実際に使用するかどうかは、HTML出力時に決めることが出来ます。
	 * ここでは、USE_MULTI_KEY_SELECT が true で、USE_SIZE(=20)以上の場合に
	 * true を返します。
	 *
	 * @og.rev 3.5.5.7 (2004/05/10) 新規作成
	 *
	 * @return  選択リストで、マルチ・キーセレクトを使用するかどうか(true:使用する)(false固定)
	 */
	public boolean useMultiSelect() {
		return false;
	}

	/**
	 * オブジェクトのキャッシュが時間切れかどうかを返します。
	 * キャッシュが時間切れ(無効)であれば、true を、有効であれば、
	 * false を返します。
	 *
	 * @og.rev 4.0.0 (2005/01/31) 新規作成
	 *
	 * @return  キャッシュが時間切れなら true
	 */
	public boolean isTimeOver() {
		return ( System.currentTimeMillis() > maxCacheTime ) ;
	}
}
